MAX_PLOTS_PER_ROW = 5
COLOR_MAJORITY = 'tab:blue'
COLOR_MINORITY = 'tab:orange'
COLOR_MIXED = 'tab:green'
COLOR_BLACK = 'black'
INACTIVE_COLOR = 'lightgrey'
RANKING_LABEL = 'rank k'
FM_TOPK_AXIS_LABEL = 'f$_m$ in top-k'
GINI_TOPK_AXIS_LABEL = 'Gini in top-k'
INEQUITY_AXIS_LABEL = 'ME'
INEQUALITY_AXIS_LABEL = 'Gini'
DPI = 300
DEFAULT_CELL_SIZE = 3
DEFAULT_FIGSIZE = (10, 5)
MINORITY_CURVE = ['get_fraction_of_minority']
TYPE_OF_DISTRIBUTION = ['pdf', 'cdf', 'ccdf']