RANDOM_NODES = "Random node sampling"
RANDOM_EDGES = "Random edge sampling"
DEGREE_RANK = "Degree Rank"
DEGREE_GROUP_RANK = "Degree Group Rank"
RANDOM_NEIGHBORS = "Random neighbor sampling"
PARTIAL_CRAWLS = "Partial crawl sampling"

DESC = "desc"
ASC = "asc"

SNSIZE = 0.01

MAX_TRIES = 20
MIN_EDGES = 2
MIN_CLASSES = 2
